import { recoverTypedSignature, signTypedData, SignTypedDataVersion } from '@metamask/eth-sig-util';
import Web3 from 'web3';
import 'whatwg-fetch';
import { JazzIcon } from './comp/JazzIcon';
import { ZWalletConnect } from './comp/ZWalletConnect';
import {
  AVAILABLE_CHAINS,
  NATIVE_PK_PREFIX,
  PAY_API_HOST,
  PAY_API_HOST_TEST,
  TX_CONFIRM_BLOCKS,
  WALLET_STORAGE_KEY_NAME,
} from './config/constants';
import { AllChains } from './data/allchain';
import { singleton } from './decorator/singleton.decorator';
import { parseUrl } from './manage/SchemeManage';
import {
  exportSecKey,
  loadInternalWallet,
  loginByEmail,
  syncWalletEnv,
  verifyPassword,
  walletPreLogin,
} from './manage/WalletManage';
import { EmailVerifySvr } from './services/EmailVerifySvr';
import { NativeSvr } from './services/NativeSvr';
import { PaySvr } from './services/PaySvr';
import { TranHistorySvr } from './services/TranHistorySvr';
import { ChainCommon } from './standards/ChainCommon';
import { ERC1155Standard } from './standards/ERC1155Standard';
import { ERC20Standard } from './standards/ERC20Standard';
import { ERC721Standard } from './standards/ERC721Standard';
import { JCStandard } from './standards/JCStandard';
import { WalletType } from './types/data.enums';
import { IChainData } from './types/data.types';
import { universalChainCb } from './util/chain.util';
import { fromTokenMinimalUnit, renderFromTokenMinimalUnit, safeNumberToBN } from './util/number.util';
import { buildLoginSignMsg, signLogin } from './util/sign.util';
import { resetAccount, resetWallet } from './api/WalletApi';

var global =
  (typeof globalThis !== 'undefined' && globalThis) ||
  (typeof self !== 'undefined' && self) ||
  (typeof global !== 'undefined' && global) ||
  {};

window.debug = true;

@singleton
export default class JCWallet {
  web3: Web3 = null;
  private wallet: any = null;
  private password: string = '111111';
  private chainSet: Set<number> = new Set();
  private chainMap: Map<number, IChainData> = new Map();
  private _currentChain: IChainData;
  public erc20Standard: ERC20Standard;
  public erc721Standard: ERC721Standard;
  public erc1155Standard: ERC1155Standard;
  public chainCommon: ChainCommon;
  public jcStandard: JCStandard;
  public nativeSvr: NativeSvr;
  public historySvr: TranHistorySvr;
  public emailVerifySvr: EmailVerifySvr;
  public paySvr: PaySvr;
  public wConnect: ZWalletConnect;
  private walletType: WalletType = WalletType.INTERNAL;
  private rpcUrl: string = '';
  public rpc: any = {};
  public nativeAccount = '';
  public env = 'dev';
  private web3Map = new Map();

  constructor({ type }: { type: number }) {
    this.nativeSvr = new NativeSvr();
    this.historySvr = new TranHistorySvr();
    this.emailVerifySvr = new EmailVerifySvr();
    this.paySvr = new PaySvr();
    this.walletType = type;
    window.jc = { wallet: this };
    this.init();
  }

  private updateChain(chain: number) {
    chain = chain || 80001;
    let data = this.chainMap.get(chain);
    if (!data) {
      throw new Error('no current chain data');
    }
    this._currentChain = data;
    this.rpcUrl = data.rpc;
    console.log(`rpc url: ${this.rpcUrl}`);
  }

  public get isInternal() {
    return this.walletType === WalletType.INTERNAL;
  }

  public get payUrl() {
    if (this.env === 'release') {
      return PAY_API_HOST;
    } else {
      return PAY_API_HOST_TEST;
    }
  }

  public get platform() {
    // @ts-ignore
    if (window.JavascriptJavaBridge) {
      console.log('regist android jsb.reflection');
      return 'android';
      // @ts-ignore
    } else if (window.JavaScriptObjCBridge) {
      return 'ios';
    }
  }

  public preLogin(channel: number, env: string = 'dev') {
    this.env = env;
    return walletPreLogin(channel);
  }

  public verifyLocalPass(pass: string) {
    return verifyPassword(pass);
  }

  public exportPrivateKey(pass: string) {
    return exportSecKey(pass);
  }

  public emailLogin(email: string, password: string) {
    return loginByEmail(email, password);
  }

  public async initInternalWallet(chain: number, pass: string, env: string = 'dev') {
    this.env = env;
    this.walletType = WalletType.INTERNAL;
    this.updateChain(chain);
    let address: string = await loadInternalWallet(pass);
    this.nativeAccount = address;
    console.log('native wallet address: ' + address);
    var start = Date.now();
    this.web3 = this.generateWeb3(chain);
    console.log(`init web3 cost: ${(Date.now() - start) / 1000}s`);
    this.erc20Standard = new ERC20Standard(this.web3);
    console.log('init Erc20Standard');
    this.erc721Standard = new ERC721Standard(this.web3);
    console.log('init Erc721Standard');
    this.erc1155Standard = new ERC1155Standard(this.web3);
    console.log('init ERC1155Standard');
    this.jcStandard = new JCStandard(this.web3);
    this.chainCommon = new ChainCommon(this.web3);
    console.log('init JCStandard');
    console.log('this.web3.eth.defaultAccount: ' + JSON.stringify(this.web3.eth.defaultAccount));
  }
  /**
   * init wallet connect
   * @returns
   */
  public async initThirdPartyWallet(chain: number, env: string = 'dev') {
    this.env = env;
    this.walletType = WalletType.THIRD_PATH;
    this.updateChain(chain);
    return new Promise(async (resolve, reject) => {
      for (const d of AllChains) {
        const id = d.id;
        this.rpc[id] = d.rpc;
      }
      if (this.wConnect) {
        this.wConnect.disconnect();
      }
      this.wConnect = new ZWalletConnect(this.rpc);
      // return this.wConnect.connect();
      this.wConnect
        .connect()
        .then(() => {
          //@ts-ignore
          this.web3 = new Web3(this.wConnect.provider);
          this.erc20Standard = new ERC20Standard(this.web3);
          this.erc721Standard = new ERC721Standard(this.web3);
          this.erc1155Standard = new ERC1155Standard(this.web3);
          this.jcStandard = new JCStandard(this.web3);
          this.chainCommon = new ChainCommon(this.web3);
          return this.web3.eth.getChainId();
        })
        .then((chainId: number) => {
          if (chainId === this._currentChain.id) {
            resolve && resolve('');
          } else {
            this.updateCurrentChain(this.currentChain.id)
              .then(() => {
                resolve && resolve('');
              })
              .catch((err) => {
                reject && reject('error change network');
              });
          }
        })
        .catch((err: Error) => {
          console.log('initThirdPartyWallet error: ' + err);
          reject && reject(err.message);
        });
    });
  }
  /**
   * init chain data
   * create local wallet data if there is no local wallet data
   * @returns
   */
  private init() {
    for (let chain of AVAILABLE_CHAINS) {
      this.chainSet.add(chain);
      if (!this.chainMap.has(chain)) {
        let data = AllChains.find((o) => o.id === chain);
        if (data) {
          this.chainMap.set(chain, data);
        }
      }
    }
    return;
  }

  get currentChain() {
    return this._currentChain;
  }

  get currentChainCfg() {
    return require(`./config/cfg_${this.currentChain.id}_${this.env}.js`);
  }

  updateCurrentChain(chainId: number) {
    return new Promise((resolve, reject) => {
      if (this.walletType === WalletType.INTERNAL) {
        const chainData = this.chainMap.get(chainId);
        if (!chainData) {
          reject && reject('chain data not found');
          return;
        }
        this._currentChain = chainData;
        this.web3.eth.setProvider(chainData.rpc);
        resolve && resolve(this.currentChain);
      } else {
        console.log('third party wallet change chain: ' + chainId);
        const chainData = this.chainMap.get(chainId);
        this.wConnect
          .addOrChangeChain(chainData)
          .then(() => {
            resolve && resolve(chainData);
          })
          .catch((err) => {
            reject && reject(err);
          });
        setTimeout(() => {
          // @ts-ignore
          jumpToWallet();
        }, 500);
      }
    });
  }

  get chainList() {
    return [...this.chainMap.values()];
  }

  public currentAccount() {
    if (this.walletType === WalletType.INTERNAL) {
      return this.wallet[0];
    } else {
      return this.wConnect.accounts[0];
    }
  }

  public generateWeb3(chainId: number, isDefault: boolean = true) {
    if (this.web3Map.has(chainId)) {
      return this.web3Map.get(chainId);
    }
    const chainData = this.chainMap.get(chainId);
    if (!chainData) {
      return null;
    }
    let web3 = new Web3(chainData.rpc);
    web3.eth.transactionConfirmationBlocks = TX_CONFIRM_BLOCKS;
    this.injectAccount(web3, isDefault);
    this.web3Map.set(chainId, web3);
    return web3;
  }

  public get currentAccAddr() {
    if (this.walletType === WalletType.INTERNAL) {
      return this.nativeAccount;
    } else {
      return this.wConnect.accounts[0];
    }
  }

  public injectAccount(web3: any, isDefault: boolean = true) {
    const nativePrivateKey = `${NATIVE_PK_PREFIX}${this.currentAccAddr.slice(2)}`;
    const acc = web3.eth.accounts.privateKeyToAccount(nativePrivateKey);
    const account = web3.eth.accounts.wallet.add(acc);
    web3.eth.defaultAccount = account.address;
    console.log('web3 account: ' + JSON.stringify(web3.eth.accounts.wallet[0]));
    if (isDefault) {
      this.wallet = web3.eth.accounts.wallet;
      web3.eth.accounts.wallet.save(this.password, WALLET_STORAGE_KEY_NAME);
    }
    return account.address;
  }

  public async generateGasShow(gas: any) {
    let chainData = this.currentChain;
    let price = await new PaySvr().getGasPrice(chainData.id);
    let ehtBN = safeNumberToBN(price).mul(safeNumberToBN(gas));
    let ethSymbol = chainData.type !== 'Testnet' ? chainData.symbol : 'ETH';
    let network = chainData.type !== 'Testnet' ? chainData.network || chainData.symbol : 'ETH';
    let fiatPrice = await new PaySvr().queryTokenPrice(network, ethSymbol);
    let priceFloat = parseFloat(fiatPrice) * 100;
    let fiatPriceBN = safeNumberToBN(priceFloat);
    let usd = fromTokenMinimalUnit(ehtBN.mul(fiatPriceBN), 20);
    let eth = fromTokenMinimalUnit(ehtBN, 18);
    return { gas, price, eth, usd };
  }

  public async fetchGasPrice() {
    let gasPrice = await this.web3.eth.getGasPrice();
    console.log('fetchGasPrice::' + gasPrice);
    let gasPriceNum = (parseInt(gasPrice) * 1.3) | 0;
    return gasPriceNum + '';
  }
  /**
   * 查询token的法币价格
   * token 必须已经在去中心化交易所上架
   * @param tokenName token名称
   * @param amount  token数量
   * @returns
   */
  public async calcTokenPrice(tokenName: string, amount: string) {
    let chainData = this.currentChain;
    let network = chainData.type !== 'Testnet' ? chainData.network || chainData.symbol : 'ETH';
    let env = this.env;
    let fiatPrice = await new PaySvr().queryTokenPrice(network, tokenName, env);
    let priceFloat = parseFloat(fiatPrice);
    return priceFloat * parseFloat(amount);
  }

  public async sendEth(to: string, amount: number | string, estimate: number) {
    let from = this.currentAccAddr;
    const amountToSend = this.web3.utils.toWei(amount + '', 'ether');
    let gas = await this.web3.eth.estimateGas({
      from,
      to,
      value: '0',
    });
    if (estimate) {
      return this.generateGasShow(gas);
    }
    const reqData = {
      from,
      to,
      gas,
      value: amountToSend,
    };
    const logData = {
      gas,
      title: 'transfer',
      details: [
        {
          address: 'eth',
          from,
          to,
          value: amountToSend,
          id: '0',
        },
      ],
    };
    return universalChainCb(logData, this.web3.eth.sendTransaction(reqData));
  }

  public async getBalance(account?: string) {
    console.log('get balance with address: ', account);
    let balance = await this.chainCommon.getBalance(account);
    return balance;
  }

  public signTypedDataV4(signObj: any) {
    const account = this.currentAccount();
    return signTypedData({
      data: signObj,
      privateKey: Buffer.from(account.privateKey.replace('0x', ''), 'hex'),
      version: SignTypedDataVersion.V4,
    });
  }

  public loginSign(nonce: string, tips: string, address?: string) {
    return new Promise((resolve, reject) => {
      const account = this.currentAccount();
      if (this.walletType === WalletType.INTERNAL) {
        const sig = signLogin(nonce, tips, account.privateKey);
        resolve && resolve(sig);
      } else {
        const signObj = buildLoginSignMsg(nonce, tips);
        this.wConnect
          .signData(signObj, account)
          .then((sig) => {
            resolve && resolve(sig);
          })
          .catch((err) => {
            reject && reject(err);
          });
        setTimeout(() => {
          // @ts-ignore
          jumpToWallet();
        }, 500);
      }
    });
  }

  public recoverTypedSignatureV4(signObj: any, signature: string) {
    return recoverTypedSignature({
      data: signObj,
      signature,
      version: SignTypedDataVersion.V4,
    });
  }

  public generateIconData(msg: string, diameter: number) {
    let icon = new JazzIcon();
    return icon.init(msg, diameter);
  }

  public formatPrice(value: string | number, decimal: number, fixed: number) {
    return renderFromTokenMinimalUnit(value, decimal, fixed);
  }

  public async erc20Info(address: string) {
    let symbol = await this.erc20Standard.getTokenSymbol(address);
    let decimal = await this.erc20Standard.getTokenDecimals(address);
    return { symbol, decimal };
  }

  public async erc20Balance(address: string, account?: string) {
    if (!account) {
      account = this.currentAccAddr;
    }
    let result = await this.erc20Standard.getBalanceOf(address, account);
    return result;
  }

  public async sendErc20(address: string, to: string, amount: string, estimate: number) {
    let from = this.currentAccAddr;
    let result = await this.erc20Standard.transfer({
      address,
      from,
      to,
      amount,
      estimate,
    });
    return result;
  }

  public async sendNFT(address: string, to: string, tokenId: string, estimate: number) {
    let from = this.currentAccAddr;
    return this.erc721Standard.transfer({
      address,
      from,
      to,
      tokenId,
      estimate,
    });
  }

  public async erc1155Balance(address: string, account: string, tokenId: string) {
    let result = await this.erc1155Standard.getBalanceOf(address, account, tokenId);
    return result;
  }

  public async erc1155Info(address: string) {}

  public async erc1155List(address: string, account?: string) {
    if (!account) {
      account = this.currentAccAddr;
    }
  }

  public async sendErc1155(address: string, to: string, tokenIds: string[], amounts: string[], estimate: number) {
    let from = this.currentAccAddr;
    return this.erc1155Standard.transferBatch({
      address,
      from,
      to,
      tokenIds,
      amounts,
      estimate,
    });
  }

  public async scanQr(title: string) {
    let result = (await this.nativeSvr.scanQRCode(title)) + '';
    console.log('scan qr code: ' + result);
    if (result && result.indexOf('://') >= 0) {
      parseUrl(result);
    }
    return result;
  }

  public async deleteAccount() {
    await resetAccount();
    return {};
  }

  public async resetWalletAddress() {
    await resetWallet();
    await syncWalletEnv();
    return {};
  }

  public saveLocalItem(key: string, value: string) {
    jsb.saveLocalStorage(key, value);
  }

  public loadLocalItem(key: string) {
    let val = jsb.loadLocalStorage(key);
    console.log('loadLocalItem:: ' + val);
    return val;
  }
}

// window.jc = window.jc || {wallet: new JCWallet()};

export * from './common/WalletEvent';
export * from './common/ZError';
export * from './lib/WalletConnect';
export * from './util/number.util';
export * from './util/wallet.util';
