import {
  checkEmailRegister,
  emailRegister,
  isEmailVerified,
  sendCode,
  verifyEmailByCode,
} from "../api/EmailApi";
import { ZError } from "../common/ZError";

import { singleton } from "../decorator/singleton.decorator";

@singleton
export class EmailVerifySvr {
  /**
   * Checks if the user's email is verified
   * @returns an object containing the verification status and the user's email
   */
  public async checkEmailVerified(): Promise<{
    verified: number;
    email: string | null;
  }> {
    let res = await isEmailVerified();
    if (res.errcode) {
      throw new ZError(res.errcode, res.errmsg); //Throw error if call to the checking function fails
    }

    const { data } = res;
    if (!data) {
      throw new Error("Invalid response from the server"); //Throw error if the expected data structure is missing
    }

    const { verified, email } = data;
    if (!email || typeof verified !== "number") {
      throw new Error("Invalid values returned from server"); //Throw error if the data doesn't have the expected types or values
    }

    return { verified, email };
  }

  /**
   * Begins the process of verifying the user's email
   * @param email - the email to be verified
   */

  public async sendEmailCode(email: string, type: number) {
    let result = await sendCode({ email, type });
    console.log(JSON.stringify(result));
    if (result.errcode) {
      throw new ZError(result.errcode, result.errmsg);
    }
    return result.data;
  }

  public async updateEmailVerify(email: string, code: string) {
    let result = await verifyEmailByCode({ email, code });
    if (result.errcode) {
      throw new ZError(result.errcode, result.errmsg);
    }
    return result.data;
  }

  public async isEmailRegister(email: string) {
    let result = await checkEmailRegister({ email });
    if (result.errcode) {
      throw new ZError(result.errcode, result.errmsg);
    }
    return result.data;
  }

  public async registByEmail(email: string, password: string, code: string) {
    password = jsb.hashSvrPass(password);
    let result = await emailRegister({ email, password, code });
    if (result.errcode) {
      throw new ZError(result.errcode, result.errmsg);
    }
    return result.data;
  }
}
