import Web3 from 'web3';
import { abiMinterFactory } from '../abis/abiUserMinterFactory';
import { GAS_BOOST, ZERO_ADDRESS } from '../config/constants';
import { universalChainCb } from '../util/chain.util';
import { abiMarket } from '../abis/abiMarket';
import { abiMall } from '../abis/abiMall';
import { abiGameMarket } from '../abis/abiGameMarket';
import { abiGameMall } from '../abis/abiGameMall';
import { abiERC721Staking } from '../abis/abiERC721Staking';
import { ZError } from '../common/ZError';
import { abiERC20 } from '../abis/abiERC20';
import { abiERC721 } from '../abis/abiERC721';
import { toBN } from '../util/number.util';
import { abiTokenMall } from '../abis/abiTokenMall';

export class JCStandard {
  private web3: Web3;

  constructor(web3: Web3) {
    this.web3 = web3;
  }

  async mintNFT({
    address,
    tokenIds,
    startTime,
    saltNonce,
    signature,
    gas,
    estimate,
  }: {
    address: string;
    tokenIds: string[];
    startTime: string;
    saltNonce: string;
    signature: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    const addressFactory = cfg.contracts.minterFactory;
    console.log('addressFactory:: ', addressFactory);
    const contract = new this.web3.eth.Contract(abiMinterFactory, addressFactory, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    if (!gas) {
      gas = await contract.methods.mintNft(address, tokenIds, startTime, saltNonce, signature).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;
    if (estimate) {
      return jc.wallet.generateGasShow(gas);
    }

    let details: any[] = [
      {
        address: addressFactory,
        from: jc.wallet.currentAccAddr,
        to: addressFactory,
      },
    ];
    for (let tokenId of tokenIds) {
      details.push({
        address,
        from: ZERO_ADDRESS,
        to: jc.wallet.currentAccAddr,
        id: tokenId,
      });
    }
    const logData = {
      gas,
      title: 'mint_nft',
      details: details,
    };
    let gasPrice = await jc.wallet.fetchGasPrice();
    return universalChainCb(
      logData,
      contract.methods.mintNft(address, tokenIds, startTime, saltNonce, signature).send({ gas, gasPrice })
    );
  }
  // begin of NFT Market
  async marketSellNFT({
    nftToken,
    currency,
    tokenId,
    price,
    amount,
    gas,
    estimate,
  }: {
    nftToken: string;
    currency: string;
    tokenId: string;
    price: string;
    amount: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressMarket = cfg.contracts.nftmarket;
    console.log('addressMarket:: ', addressMarket);
    const contract = new this.web3.eth.Contract(abiMarket, addressMarket, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    const tokenInstance = new this.web3.eth.Contract(abiERC721, nftToken, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let approved = await tokenInstance.methods.getApproved(tokenId);
    let gasApprove = 0;
    let gasPrice = await jc.wallet.fetchGasPrice();
    if (approved != addressMarket) {
      gasApprove = await tokenInstance.methods.approve(addressMarket, tokenId).estimateGas();
      gasApprove = (gasApprove * GAS_BOOST) | 1;
      await tokenInstance.methods.approve(addressMarket, tokenId).send({ gas: gasApprove, gasPrice });
    }

    if (!gas) {
      gas = await contract.methods.sell(nftToken, currency, tokenId, price, amount).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;

    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }

    let details: any = [
      {
        address: addressMarket,
        from: jc.wallet.currentAccAddr,
        to: addressMarket,
      },
      {
        address: nftToken,
        from: jc.wallet.currentAccAddr,
        to: addressMarket,
        id: tokenId,
      },
    ];
    const logData = {
      gas,
      title: 'market_sell',
      details: details,
    };

    return universalChainCb(
      logData,
      contract.methods.sell(nftToken, currency, tokenId, price, amount).send({ gas, gasPrice })
    );
  }

  async marketUpdatePrice({
    orderId,
    price,
    gas,
    estimate,
  }: {
    orderId: string;
    price: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressMarket = cfg.contracts.nftmarket;
    console.log('addressMarket:: ', addressMarket);
    const contract = new this.web3.eth.Contract(abiMarket, addressMarket, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    if (!gas) {
      gas = await contract.methods.updatePrice(orderId, price).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;
    if (estimate) {
      return jc.wallet.generateGasShow(gas);
    }

    let details = [
      {
        address: addressMarket,
        from: jc.wallet.currentAccAddr,
        to: addressMarket,
      },
    ];
    const logData = {
      gas,
      title: 'market_update_price',
      details: details,
    };
    let gasPrice = await jc.wallet.fetchGasPrice();
    return universalChainCb(logData, contract.methods.updatePrice(orderId, price).send({ gas, gasPrice }));
  }

  async marketCancelOrder({ orderId, gas, estimate }: { orderId: string; gas?: number; estimate: number }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressMarket = cfg.contracts.nftmarket;
    console.log('addressMarket:: ', addressMarket);
    const contract = new this.web3.eth.Contract(abiMarket, addressMarket, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    if (!gas) {
      gas = await contract.methods.cancelOrder(orderId).estimateGas();
      // gas = cfg.gasInfo.marketCancelOrder;
    }
    gas = (gas * GAS_BOOST) | 1;
    if (estimate) {
      return jc.wallet.generateGasShow(gas);
    }
    let orderInfo = await contract.methods.orderInfos(orderId).call();
    if (!orderInfo.orderId) {
      throw new ZError(100, `order info with id: ${orderId} not found`);
    }
    let details = [
      {
        address: addressMarket,
        from: jc.wallet.currentAccAddr,
        to: addressMarket,
      },
      {
        address: orderInfo.nftToken,
        from: addressMarket,
        to: jc.wallet.currentAccAddr,
        id: orderInfo.tokenId,
      },
    ];
    let gasPrice = await jc.wallet.fetchGasPrice();
    const logData = {
      gas,
      title: 'market_cancel_order',
      details: details,
    };
    return universalChainCb(logData, contract.methods.cancelOrder(orderId).send({ gas, gasPrice }));
  }

  async marketBuy({
    orderId,
    gas,
    price,
    estimate,
  }: {
    orderId: string;
    price: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressMarket = cfg.contracts.nftmarket;
    console.log('addressMarket:: ', addressMarket);
    const contract = new this.web3.eth.Contract(abiMarket, addressMarket, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let orderInfo = await contract.methods.orderInfos(orderId).call();
    if (!orderInfo.orderId) {
      throw new ZError(100, `order info with id: ${orderId} not found`);
    }
    if (!toBN(price).eq(toBN(orderInfo.price))) {
      throw new ZError(101, `price must equal to ${orderInfo.price}`);
    }
    const tokenInstance = new this.web3.eth.Contract(abiERC20, orderInfo.currency, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let approved = await tokenInstance.methods.allowance(jc.wallet.currentAccAddr, addressMarket).call();
    console.log('approved:: ', approved);
    let gasApprove = 0;
    let gasPrice = await jc.wallet.fetchGasPrice();
    if (toBN(approved).lt(toBN(price))) {
      gasApprove = await tokenInstance.methods.approve(addressMarket, price).estimateGas();
      gasApprove = (gasApprove * GAS_BOOST) | 1;
      await tokenInstance.methods.approve(addressMarket, price).send({ gas: gasApprove, gasPrice });
    }
    if (!gas) {
      gas = await contract.methods.buy(orderId, price).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;
    // begin of increase allowance

    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }
    let details = [
      {
        address: addressMarket,
        from: jc.wallet.currentAccAddr,
        to: addressMarket,
      },
      {
        address: orderInfo.nftToken,
        from: addressMarket,
        to: jc.wallet.currentAccAddr,
        id: orderInfo.tokenId,
        value: orderInfo.amount,
      },
      {
        address: orderInfo.currency,
        from: jc.wallet.currentAccAddr,
        to: addressMarket,
        id: orderInfo.price,
      },
    ];
    const logData = {
      gas,
      title: 'market_buy',
      details: details,
    };
    return universalChainCb(logData, contract.methods.buy(orderId, price).send({ gas, gasPrice }));
  }

  async marketOrderInfo(orderId: string) {
    const cfg = jc.wallet.currentChainCfg;
    const addressMarket = cfg.contracts.nftmarket;
    console.log('addressMarket:: ', addressMarket);
    const contract = new this.web3.eth.Contract(abiMarket, addressMarket, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let orderInfo = await contract.methods.orderInfos(orderId).call();
    return orderInfo;
  }
  // end of NFT Market

  // begin of game item mall
  async gameMallBuy({
    orderId,
    currency,
    price,
    startTime,
    saltNonce,
    signature,
    gas,
    estimate,
  }: {
    orderId: string;
    currency: string;
    price: string;
    startTime: string;
    saltNonce: string;
    signature: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressGameMall = cfg.contracts.gamemall;
    console.log('addressGameMall:: ', addressGameMall);
    const contract = new this.web3.eth.Contract(abiGameMall, addressGameMall, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    const tokenInstance = new this.web3.eth.Contract(abiERC20, currency, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let approved = await tokenInstance.methods.allowance(jc.wallet.currentAccAddr, addressGameMall).call();
    console.log('approved:: ', approved);
    let gasApprove = 0;
    let gasPrice = await jc.wallet.fetchGasPrice();
    if (toBN(approved).lt(toBN(price))) {
      gasApprove = await tokenInstance.methods.approve(addressGameMall, price).estimateGas();
      gasApprove = (gasApprove * GAS_BOOST) | 1;
      console.log('gasApprove:: ', gasApprove);
      await tokenInstance.methods.approve(addressGameMall, price).send({ gas: gasApprove, gasPrice });
    }

    if (!gas) {
      gas = await contract.methods.buy(orderId, currency, price, startTime, saltNonce, signature).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;
    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }
    let feeAddress = await contract.methods.feeToAddress().call();
    let details = [
      {
        address: addressGameMall,
        from: jc.wallet.currentAccAddr,
        to: addressGameMall,
      },
      {
        address: currency,
        from: jc.wallet.currentAccAddr,
        to: feeAddress,
        value: price,
      },
    ];
    const logData = {
      gas,
      title: 'game_mall_buy',
      details: details,
    };

    return universalChainCb(
      logData,
      contract.methods.buy(orderId, currency, price, startTime, saltNonce, signature).send({ gas, gasPrice })
    );
  }
  // end of game item mall

  // begin of game item market
  async gameMarketBuy({
    orderId,
    seller,
    currency,
    price,
    startTime,
    saltNonce,
    signature,
    gas,
    estimate,
  }: {
    orderId: string;
    seller: string;
    currency: string;
    price: string;
    startTime: string;
    saltNonce: string;
    signature: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressGameMarket = cfg.contracts.gamemarket;
    console.log('addressGameMarket:: ', addressGameMarket);
    const contract = new this.web3.eth.Contract(abiGameMarket, addressGameMarket, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    const tokenInstance = new this.web3.eth.Contract(abiERC20, currency, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let approved = await tokenInstance.methods.allowance(jc.wallet.currentAccAddr, addressGameMarket).call();
    console.log('approved:: ', approved);
    let gasApprove = 0;
    let gasPrice = await jc.wallet.fetchGasPrice();
    if (toBN(approved).lt(toBN(price))) {
      gasApprove = await tokenInstance.methods.approve(addressGameMarket, price).estimateGas();
      gasApprove = (gasApprove * GAS_BOOST) | 1;
      await tokenInstance.methods.approve(addressGameMarket, price).send({ gas: gasApprove, gasPrice });
    }
    console.log('gasApprove: ' + gasApprove);
    if (!gas) {
      gas = await contract.methods.buy(orderId, seller, currency, price, startTime, saltNonce, signature).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;
    console.log('gas: ' + gas);

    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }
    let feeAddress = await contract.methods.feeToAddress().call();
    let details = [
      {
        address: addressGameMarket,
        from: jc.wallet.currentAccAddr,
        to: addressGameMarket,
      },
      {
        address: currency,
        from: jc.wallet.currentAccAddr,
        to: feeAddress,
        value: price,
      },
    ];
    const logData = {
      gas,
      title: 'game_market_buy',
      details: details,
    };
    return universalChainCb(
      logData,
      contract.methods.buy(orderId, seller, currency, price, startTime, saltNonce, signature).send({ gas, gasPrice })
    );
  }
  // end of game item market
  // begin of token Mall
  /**
   *
   * @param values: [orderId, price, startTime, saltNonce]
   * @returns
   */
  async buyTokenWithErc20({ currency, amount, estimate }: { currency: string; amount: string; estimate: number }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressMall = cfg.contracts.tokenmall;
    console.log('addressMall:: ', addressMall);
    const contract = new this.web3.eth.Contract(abiTokenMall, addressMall, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    const tokenInstance = new this.web3.eth.Contract(abiERC20, currency, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let approved = await tokenInstance.methods.allowance(jc.wallet.currentAccAddr, addressMall).call();
    console.log('approved:: ', approved);
    let gasPrice = await jc.wallet.fetchGasPrice();
    console.log('gas price: ' + gasPrice);
    let gasApprove = 0;
    if (toBN(approved).lt(toBN(amount))) {
      gasApprove = await tokenInstance.methods.approve(addressMall, amount).estimateGas();
      gasApprove = (gasApprove * GAS_BOOST) | 1;
      await tokenInstance.methods.approve(addressMall, amount).send({ gas: gasApprove, gasPrice });
    }
    let gas = await contract.methods.buyToken(currency, amount).estimateGas();
    gas = (gas * GAS_BOOST) | 1;
    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }

    let details: any = [
      {
        address: addressMall,
        from: jc.wallet.currentAccAddr,
        to: addressMall,
      },
      {
        address: currency,
        from: jc.wallet.currentAccAddr,
        to: addressMall,
        value: amount,
      },
    ];
    let tokenAddress = await contract.methods.tokenAddress().call();
    let currencyDecimals = await tokenInstance.methods.decimals().call();
    let price = await contract.methods.prices(currency).call();
    let tokenAmount = toBN(price)
      .mul(toBN(amount))
      .mul(toBN(10).pow(toBN(10).sub(toBN(currencyDecimals))))
      .div(toBN('1000000'));
    console.log(`buyTokenWithErc20: currency: ${currency}, amount: ${amount}, tokenAmount: ${tokenAmount}`);
    details.push({
      address: tokenAddress,
      from: addressMall,
      to: jc.wallet.currentAccAddr,
      value: tokenAmount,
    });
    const logData = {
      gas,
      title: 'mall_buy_token',
      details: details,
    };
    return universalChainCb(logData, contract.methods.buyToken(currency, amount).send({ gas, gasPrice }));
  }
  // end of token Mall

  // begin of NFT Mall
  /**
   *
   * @param values: [orderId, price, startTime, saltNonce]
   * @returns
   */
  async mallBuy({
    currency,
    addresses,
    ids,
    amounts,
    values,
    signature,
    gas,
    estimate,
  }: {
    currency: string;
    addresses: string[];
    ids: string[];
    amounts: string[];
    values: string[];
    signature: string;
    gas?: number;
    estimate: number;
  }) {
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressMall = cfg.contracts.nftmall;
    console.log('addressMall:: ', addressMall);
    const contract = new this.web3.eth.Contract(abiMall, addressMall, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    const tokenInstance = new this.web3.eth.Contract(abiERC20, currency, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let approved = await tokenInstance.methods.allowance(jc.wallet.currentAccAddr, addressMall).call();
    console.log('approved:: ', approved);
    let gasPrice = await jc.wallet.fetchGasPrice();
    console.log('gas price: ' + gasPrice);
    let gasApprove = 0;
    if (toBN(approved).lt(toBN(values[1]))) {
      gasApprove = await tokenInstance.methods.approve(addressMall, values[1]).estimateGas();
      gasApprove = (gasApprove * GAS_BOOST) | 1;
      await tokenInstance.methods.approve(addressMall, values[1]).send({ gas: gasApprove, gasPrice });
    }
    if (!gas) {
      gas = await contract.methods.buyNFT(currency, addresses, ids, amounts, values, signature).estimateGas();
    }
    gas = (gas * GAS_BOOST) | 1;

    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }

    let details: any = [
      {
        address: addressMall,
        from: jc.wallet.currentAccAddr,
        to: addressMall,
      },
      {
        address: currency,
        from: jc.wallet.currentAccAddr,
        to: addressMall,
        value: values[1],
      },
    ];
    for (let i = 0; i < addresses.length; i++) {
      details.push({
        address: addresses[i],
        from: jc.wallet.currentAccAddr,
        to: jc.wallet.currentAccAddr,
        value: values[i],
        id: ids[i],
      });
    }
    const logData = {
      gas,
      title: 'mall_buy_nft',
      details: details,
    };
    return universalChainCb(
      logData,
      contract.methods.buyNFT(currency, addresses, ids, amounts, values, signature).send({ gas, gasPrice })
    );
  }
  // end of NFT Mall

  // begin of staking
  async stakeNfts({
    nfts,
    tokenIds,
    staketimes,
    gas,
    estimate,
  }: {
    nfts: string[];
    tokenIds: string[];
    staketimes: string[];
    gas?: number;
    estimate: number;
  }) {
    if (nfts.length != tokenIds.length || nfts.length != staketimes.length) {
      throw new ZError(100, 'nfts, tokenIds, staketimes must have same length');
    }
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressStake = cfg.contracts.nftStaking;
    console.log('addressStake:: ', addressStake);
    const contract = new this.web3.eth.Contract(abiERC721Staking, addressStake, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let nftSet = new Set(nfts);
    let gasPrice = await jc.wallet.fetchGasPrice();
    let gasApprove = 0;
    for (let nft of nftSet) {
      let tokenInstance = new this.web3.eth.Contract(abiERC721, nft, {
        //@ts-ignore
        from: jc.wallet.currentAccAddr,
      });
      let approved = await tokenInstance.methods.isApprovedForAll(jc.wallet.currentAccAddr, addressStake).call();
      console.log('approved:: ', approved);
      if (!approved) {
        gasApprove += await tokenInstance.methods.setApprovalForAll(addressStake, true).estimateGas();
        gasApprove = (gasApprove * GAS_BOOST) | 1;
        await tokenInstance.methods.setApprovalForAll(addressStake, true).send({ gas: gasApprove, gasPrice });
      }
    }
    if (!gas) {
      gas = await contract.methods.stake(nfts, tokenIds, staketimes).estimateGas();
      gas = (gas * GAS_BOOST) | 1;
    }
    if (estimate) {
      return jc.wallet.generateGasShow(gas + gasApprove);
    }
    let details: any = [
      {
        address: addressStake,
        from: jc.wallet.currentAccAddr,
        to: addressStake,
      },
    ];
    for (let i = 0; i < nfts.length; i++) {
      details.push({
        address: nfts[i],
        from: jc.wallet.currentAccAddr,
        to: addressStake,
        value: 1,
        id: tokenIds[i],
      });
    }
    const logData = {
      gas,
      title: 'stake_nft',
      details: details,
    };
    return universalChainCb(logData, contract.methods.stake(nfts, tokenIds, staketimes).send({ gas, gasPrice }));
  }

  async redeemNfts({
    nfts,
    tokenIds,
    gas,
    estimate,
  }: {
    nfts: string[];
    tokenIds: string[];
    gas?: number;
    estimate: number;
  }) {
    if (nfts.length != tokenIds.length) {
      throw new ZError(100, 'nfts, tokenIds must have same length');
    }
    const cfg = jc.wallet.currentChainCfg;
    estimate = parseInt(estimate + '');
    const addressStake = cfg.contracts.nftStaking;
    console.log('addressStake:: ', addressStake);
    const contract = new this.web3.eth.Contract(abiERC721Staking, addressStake, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    if (!gas) {
      gas = await contract.methods.redeem(nfts, tokenIds).estimateGas();
      gas = (gas * GAS_BOOST) | 1;
    }
    if (estimate) {
      return jc.wallet.generateGasShow(gas);
    }
    let gasPrice = await jc.wallet.fetchGasPrice();
    let details: any = [
      {
        address: addressStake,
        from: jc.wallet.currentAccAddr,
        to: addressStake,
      },
    ];
    for (let i = 0; i < nfts.length; i++) {
      details.push({
        address: nfts[i],
        from: addressStake,
        to: jc.wallet.currentAccAddr,
        value: 1,
        id: tokenIds[i],
      });
    }
    const logData = {
      gas,
      title: 'redeem_nft',
      details: details,
    };
    return universalChainCb(logData, contract.methods.redeem(nfts, tokenIds).send({ gas, gasPrice }));
  }

  async nftStakeInfo({ nft, tokenId }: { nft: string; tokenId: string }) {
    const cfg = jc.wallet.currentChainCfg;
    const addressStake = cfg.contracts.nftStaking;
    console.log('addressStake:: ', addressStake);
    const contract = new this.web3.eth.Contract(abiERC721Staking, addressStake, {
      //@ts-ignore
      from: jc.wallet.currentAccAddr,
    });
    let stakeInfo = await contract.methods.stakingMap(nft, tokenId).call();
    return stakeInfo;
  }
  // end of staking
}
