import { BASE_TOKEN_URI, TX_CONFIRM_BLOCKS } from "../config/constants";
import { LoggerQueue } from "../queue/record.queue";
import { toBN } from "web3-utils";

export const UNKNOW = "unknow";

/**
 * change price with customer decimals to bigNum with 18 decimals
 * @param {number} price
 * @param {number} decimals
 * @return {string}
 */
export function parsePrice(price: number, decimals: number): string {
  const n = 19 - decimals;
  return price + new Array(n).join("0");
}

/**
 * format price with customer decimals to string for display
 * @param {number | string} price
 * @param {number} decimals
 * @param {number} fixed
 * @return {number | string}
 */
export function formatPrice(
  price: number | string,
  decimals?: number,
  fixed = 2
): string | number {
  if (!decimals) {
    return price;
  }
  let str = price + "";
  const length = str.length;
  str = str.padStart(decimals, "0");
  if (decimals >= length) {
    str = "0." + str;
  } else {
    const pos = length - decimals;
    str = str.slice(0, pos) + "." + str.slice(pos);
  }
  str = str.slice(0, str.lastIndexOf(".") + fixed + 1);
  return str;
  // return str.replace(/0+$/, '').replace(/\.+$/, '')
}

/**
 * number to hex string
 * @param {number} chainId
 * @return {string}
 */
export function toHexChainId(chainId: number) {
  return "0x" + chainId.toString(16);
}

export function getTypeByAddress(address: string) {
  const cfg = jc.wallet.currentChainCfg;
  let type = "erc721";
  if (cfg && cfg.tokens) {
    for (let token of cfg.tokens) {
      if (token.address.toLowerCase() === address.toLowerCase()) {
        type = token.type;
      }
    }
  }
  return { type };
}

export function getAddressByName(typeName: string) {
  const cfg = jc.wallet.currentChainCfg;
  let address = "";
  if (cfg && cfg.tokens) {
    let result = cfg.tokens.find(
      (token) => token.name === typeName || token.symbol === typeName
    );
    if (result) {
      address = result.address;
    }
  }
  return address;
}

export async function getJCErc721Info(tokenId: string) {
  const url = `${BASE_TOKEN_URI}${tokenId}`;
  return fetch(url).then((response) => {
    return response.json();
  });
}

export function universalChainCb(reqData: any, req: any) {
  return req
    .on("presend", function (dataObj: any) {
      if (jc.wallet.isInternal) {
        console.log("before send tran: ", JSON.stringify(dataObj));
        let gasPrice;
        if (
          !dataObj.gasPrice &&
          dataObj.maxPriorityFeePerGas &&
          dataObj.maxFeePerGas
        ) {
          gasPrice = toBN(dataObj.maxFeePerGas)
            .sub(toBN(dataObj.maxPriorityFeePerGas))
            .div(toBN(2))
            .mul(toBN(1000000000));
        } else if (dataObj.gasPrice) {
          gasPrice = toBN(dataObj.gasPrice);
        }
        reqData.gas = toBN(dataObj.gas).toString();
        reqData.gasPrice = gasPrice.toString();
        reqData.transactionHash = dataObj.transactionHash;
        reqData.chain = jc.wallet.currentChain.id;
        reqData.startTime = Date.now();
        new LoggerQueue().addLog(reqData);
      }
    })
    .on("transactionHash", function (hash: any) {
      console.log(`remote transactionHash: ${hash}`);
    })
    .on("confirmation", function (confirmationNumber, receipt) {
      console.log(
        `confirmation:: confirmationNumber: ${confirmationNumber} , receipt: ${receipt}`
      );
      if (jc.wallet.isInternal && confirmationNumber >= TX_CONFIRM_BLOCKS) {
        reqData.status = 2;
        reqData.confirmTime = Date.now();
        new LoggerQueue().addLog(reqData);
      }
    })
    .on("receipt", function (receipt) {
      console.log(receipt);
      if (jc.wallet.isInternal) {
        reqData.status = receipt.status ? 1 : 10;
        reqData.blockNumber = receipt.blockNumber;
        reqData.blockHash = receipt.blockHash;
        reqData.gas = receipt.gasUsed;
        new LoggerQueue().addLog(reqData);
      }
      return Promise.resolve(receipt);
    })
    .on("error", function (error, receipt) {
      console.log(error);
      if (jc.wallet.isInternal) {
        reqData.status = 11;
        new LoggerQueue().addLog(reqData);
      }
      return Promise.reject(error);
    });
}
